import tkinter
from tkinter import ttk
import logging
import jtk
import random
import registers
import alu
import crawler
from crawler import Crawler

INITIAL_OPCODE = '8'

class Cpu:
    def __init__(self, canvas, x, y, sep):
        self.canvas = canvas
        self.r = r = registers.RegisterFile(canvas, x, y, False)

        self.a = a = alu.Alu(canvas, x+sep, y, "label")
        self.ops = {'C': '0', 'D': '1', '8': '2'}      # translates opcode field to aluOp

## Crawler connecting read data 1 output of register file to ALU upper input
## make a Label and a line and use them to make a Crawler
## then connect the Crawler to its two endpoints
## omit done method so it takes its default value of None
        outputObject = jtk.Label(canvas)
        line = canvas.create_line(r.getRightX(), a.getLeftY(0), a.getLeftX(), a.getLeftY(0),
                                         width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.rd1crl = crawler.Crawler(canvas, line, outputObject)
        r.setRD1(self.rd1crl)           # initial endpoint of Crawler object
        a.setX(self.rd1crl)             # final endpoint of Crawler object

## Crawler connecting read data 2 output of register file to nowhere
        outputObject = jtk.Label(canvas)
        line = canvas.create_line(r.getRightX(), a.getLeftY(1), r.getRightX()+70, a.getLeftY(1),
                                         width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.rd2crl = crawler.Crawler(canvas, line, outputObject)
        r.setRD2(self.rd2crl)       # initial
#        a.setY(self.rd2crl)         # final

## Crawler connecting write data input of register file to ALU output
        inputObject = jtk.Label(canvas)
        line = canvas.create_line(a.getRightX(), a.getRightY(1), a.getRightX() + 200, a.getRightY(1),
                                   a.getRightX() + 200, a.getRightY(1) + 450,
                                   r.getLeftX() - 100, a.getRightY(1) + 450,
                                   r.getLeftX() - 100, r.getLeftY(3),
                                   r.getLeftX(), r.getLeftY(3),
                                         width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.wdcrl = crawler.Crawler(canvas, line, inputObject)
        r.setWD(self.wdcrl)
        a.setA(self.wdcrl)
## the Alu's zero output bit
        self.zval = jtk.Label(self.canvas.master)
        a.setZ(jtk.TextLine.ptLen(canvas, (a.getRightX(), a.getRightY(0)), 100, 1, 'w', self.zval, 15))

##      create the instruction register out of rectangles and their contents
        tlx = r.getLeftX() - 400         ## top left corner of instruction boxes
        tly = r.getLeftY(0) + 100        ## top left corner of instruction boxes
        canvas.create_rectangle(tlx, tly, tlx+50, tly+50, fill='white')
        cbv = list(self.ops.keys())
        cbv.sort()                  # sort does not return a value, so we need a variable here
        self.cbb = ttk.Combobox(canvas, width=3, values=cbv, state="readonly", takefocus=False)
        self.cbb.bind("<<ComboboxSelected>>", self.doOp)
        self.cbb.set(INITIAL_OPCODE)
        self.doOp()                                 # make ALU correspond to opcode
        canvas.create_window(tlx+25, tly+25, window=self.cbb)                   # opcode is ComboBox
        canvas.create_rectangle(tlx+50, tly, tlx+100, tly+50, fill='white') 
        rs = canvas.create_text(tlx+75, tly+25, justify="center", text=random.randint(0,7))      # rs is random
        canvas.create_rectangle(tlx+100, tly, tlx+150, tly+50, fill='white')
        rt = canvas.create_text(tlx+125, tly+25, text=random.randint(0,7))              #rt is random
        canvas.create_rectangle(tlx+150, tly, tlx+300, tly+50, fill='white') 
        imo = canvas.create_text(tlx+175, tly+25, text=random.randint(-100, 100))       # immediate operand is random

## Crawler connecting rs field of instruction register to read register 1 input of register file
        outputObject = jtk.Label(canvas)                                                ## rs
        line = canvas.create_line(tlx+75, tly, tlx+75, r.getLeftY(0)-25, r.getLeftX(), r.getLeftY(0)-25,
                                  width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.rscrl = crawler.Crawler(canvas, line, outputObject)
        self.r.setRR1(self.rscrl)
        self.rscrl.set(canvas.itemcget(rs, "text"))

## Crawler connecting rt field of instruction register to read register 2 input of register file
        outputObject = jtk.Label(canvas)
        line = canvas.create_line(tlx+125, tly, tlx+125, r.getLeftY(1)-60,
                                  r.getLeftX(), r.getLeftY(1)-60,
                                  width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.rtcrl1 = crawler.Crawler(canvas, line, outputObject)
        self.rtcrl1.set(canvas.itemcget(rt, "text"))

## Crawler connecting read register 2 input of register file to write register input of register file
        outputObject = jtk.Label(canvas)
        line = canvas.create_line(r.getLeftX()-35, r.getLeftY(1)-80, r.getLeftX()-35, r.getLeftY(2),
                                  r.getLeftX(), r.getLeftY(2),
                                  width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.rtcrl2 = crawler.Crawler(canvas, line, outputObject)
        self.r.setWR(self.rtcrl2)
#        self.rtcrl2.set(canvas.itemcget(rt, "text"))
# ordinary line connecting rt field of instruction register to read register 2 input of register file
        canvas.create_line(r.getLeftX()-40, r.getLeftY(1)-60, r.getLeftX(), r.getLeftY(1)-60,
                           width=40, joinstyle="miter", arrow=tkinter.LAST)
        self.r.setRR2(self.rtcrl1)  # use rt field to set RR2;

## Crawler connecting immediate operand field of instruction register to lower input of ALU
        outputObject = jtk.Label(canvas)
        line = canvas.create_line(tlx+175, tly+51, tlx+175, r.getLeftY(2)+200,
                                  r.getRightX()+150, r.getLeftY(2)+200, r.getRightX()+150, a.getLeftY(1),
                                  a.getLeftX(), a.getLeftY(1),
                                         width=40, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
        self.imcrl = crawler.Crawler(canvas, line, outputObject)
        a.setY(self.imcrl)
        self.imcrl.set(canvas.itemcget(imo, "text"))

        # winfo_reqwidth() works; winfo_width() will not work because the window has not yet been rendered
        self.b = ttk.Button(self.canvas.master, text="start", command = self.pressed)
        self.canvas.create_window(self.canvas.winfo_reqwidth()/2, 15, window=self.b)

        self.aluWaiting = 2
        self.writeWaiting = 2
        self.startem()
        # end of Cpu.__init__()
        #**********************************************************
        #**********************************************************

    def doOp(self, event=None):
        op = self.cbb.get()
        opval = self.ops[op]
        self.a.op.set(self.a.opTrans(opval))

    def pressed(self):                              ## start/stop button
        if Crawler.going:
            self.b["text"] = "start"
            Crawler.going = False
        else:
            self.b["text"] = "stop"
            Crawler.going = True

## in what follows, the parameter to the start method is the method to be invoked at completion of the started action
# start first three Crawlers
    def startem(self):
        self.rscrl.start(self.rs_cmd)       # start rscrl; invoke rs_cmd on completion
        self.rtcrl1.start(self.rt_cmd)      # start rtcrl; invoke rt_cmd on completion
        self.imcrl.start(self.alu_cmd)      # start imcrl; invoke alu_cmd on completion
        Crawler.going = False

    def rs_cmd(self):                       # RD1 goes to alu
        self.r.doRD1()
        self.rd1crl.start(self.alu_cmd)

    def rt_cmd(self):
        self.r.doRD2()
        self.rtcrl2.set(self.rtcrl1.get())
        self.rd2crl.start(None)             # this chain stops here; RD2 is not used
        self.rtcrl2.start(self.wb_cmd)      # rt also specifies write register

    def alu_cmd(self):
        self.aluWaiting -= 1
        if self.aluWaiting == 0:            # both values have arrived
            self.a.doOp()
            self.wdcrl.start(self.wb_cmd)

    def wb_cmd(self):
        self.writeWaiting -= 1
        if self.writeWaiting == 0:          # register number and new value have both arrived
            self.r.doButton()






if __name__ == "__main__":
    TopLeft_X, TopLeft_Y = 450, 100
    Separation = 500
    original_width, original_height = 1440, 900      # screen dimensions of my laptop

    class App(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            self.canvas = tkinter.Canvas(self, width=original_width, height=original_height)
            Cpu(self.canvas, TopLeft_X, TopLeft_Y, Separation)
            self.canvas.addtag_all("all")    # tag all widgets in canvas to be resized along with it
            new_width, new_height = min(self.winfo_screenwidth(), original_width), min(self.winfo_screenheight(), original_height)
            sf = self.get_scale_factor(new_width, new_height)
            # resize the canvas 
            self.canvas.config(width=new_width, height=new_height)
            parent.minsize(new_width, new_height)
            parent.maxsize(new_width, new_height)
            # rescale all the objects tagged with the "all" tag
            self.canvas.scale("all", 0, 0, sf, sf)
            self.canvas.pack()

        def get_scale_factor(self, nw, nh):
            wscale, hscale = min(nw / float(original_width), 1.0), min(nh / float(original_height), 1.0)
            scale_factor = min(wscale, hscale)
            logging.debug(scale_factor)
            return scale_factor

    def main():
        root = tkinter.Tk()
        root.title("Model CPU for addi, andi, and ori Instructions")
        a = App(root)
        a.pack()
        root.mainloop()

    main()
