import tkinter
from tkinter import ttk
import jtk
import logging
import random

class CtrlReg(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent, padding="3 3 3 3")
        ttk.Label(self, text="Control Register").grid(column=0, row=0, padx=20)
        ttk.Entry(self, width=30, state="disabled").grid(column=1, row=0)   # empty
        self.bitOne = jtk.BitField(self, self.master.doBad, postOK=lambda e: 'pass', readonly=True)
        self.bitOne.configure(exportselection=False)
        self.bitZero = jtk.BitField(self, self.master.doBad, postOK=lambda e: 'pass', readonly=True)
        self.bitZero.configure(exportselection=False)
        self.setReady(True)
        self.bitOne.grid(column=2, row=0, padx=0, pady=0)
        self.bitZero.grid(column=3, row=0, padx=0, pady=0)

##    def postBitsOK(self):
##        pass
        
    def isReady(self):  ## returns  Boolean
        return self.bitZero.get() == '1'
        
    def setReady(self, willBeReady):
        self.bitZero.state(["!readonly"])
        if willBeReady:
            self.bitZero.unvalidatedSetStr('1')
        else:
            self.bitZero.unvalidatedSetStr('0')
        self.bitZero.state(["readonly"])
        
    def getEnable(self):  ## returns  integer
        return int(self.bitOne.get())

class DataReg(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent, padding="3 3 3 3")
        ttk.Label(self, text="Data Register    ").grid(column=0, row=0, padx=20)
        ttk.Entry(self, width=25, state="disabled").grid(column=1, row=0)
        self.data = ttk.Entry(self, width=8, state="readonly", exportselection=False)
        self.setData('0')
        self.data.grid(column=2, row=0)

    def getData(self):
        val = self.data.get()
        if len(val) > 0:
            return int(val)
        else:
            return None

    def setData(self, newData):
        self.data.state(["!readonly"])
        self.data.delete(0, "end")
        self.data.insert(0, newData)
        self.data.state(["readonly"])

class ButtonPanel(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        ttk.Label(self, text="Program").grid(column=0, row=0, padx=10)
        self.data = ttk.Entry(self, width=27, takefocus=False)
        self.data.insert(0, "ABCDEFGHIJKLMNOPQRSTUVWXYZ")
        self.data.select_range(0, 1)
        self.data.state(["readonly"])
        button = ttk.Button(self, text="Write Data", command=self.master.write)
        self.cb = jtk.Checkbutton(self, True, text="leave on")
        self.data.grid(column=1, row=0, padx=10)
        button.grid(column=2, row=0, padx=10)
        self.cb.grid(column=3, row=0, padx=10)
        self.data.bind('<ButtonPress>', lambda e: 'break')
        self.data.bind('<Motion>', lambda e: 'break')
        
    def trim(self, string):  ## get rid of leading zeroes but not entire string
            i = 0
            for i in range(len(string) - 1):
                if string[i] != '0': break
                i += 1
            return string[i:]

    def get(self):
        return self.data.selection_get()

    def advance(self):
        stop = self.data.index("sel.last")
        end = self.data.index("end")
        if stop == end:
            self.data.select_range(0, 1)
        else:
            self.data.select_range(stop, stop+1)
class OutputPanel(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        ttk.Label(self, text="Output").grid(column=0, row=0, padx=20)
        self.display = tkinter.Text(self, width=50, height=10, wrap="char", takefocus="False",
                                         state='disabled', exportselection=False)
        self.display.grid(column=1, row=0)

    def write(self, ch):
        self.display.config(state="normal")
        self.display.insert("end", ch)    ## write character on output page
        self.display.config(state="disabled")
        

    
class App(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)
        logging.basicConfig(level=logging.WARNING)
        self.dspCtrl = CtrlReg(self)
        self.dspData = DataReg(self)
        self.bp = ButtonPanel(self)
        self.opp = OutputPanel(self)
        self.dspCtrl.pack(padx=10)
        self.dspData.pack(padx=10)
        self.bp.pack()
        self.opp.pack(padx=10)

    def write(self):
        char = self.bp.get()                ## gets selected character in alphabet string
        oc = ord(char)
        self.bp.advance()                   ## move to next character or start over
        self.dspData.setData(oc)            ## char code --> display data register
        if self.dspCtrl.isReady():
            self.opp.write(char)
            if not self.bp.cb.isSelected():
                self.dspCtrl.setReady(False)
                delay = random.randint(0, 4) * 250     ## 0 - 4 quarters of a second
                self.after(2000+delay, self.reset)

    def reset(self):
        self.dspCtrl.setReady(True)
    
    def doBad(self):
        self.bell()

if __name__ == "__main__":
    root = tkinter.Tk()
    root.title("The Ready Bit")
    a = App(root)
    a.pack()
    root.mainloop()
