"""
Here are the magic numbers in the MIPS address space.
They are expressed as hexadecimal, decimal, sum of powers of 2, fraction of the whole,
  and size on a scale where 16 inches is the entire space.


0x0040-0000 = 4194304 = 2^22 = 1/1024  = 1/64 inch  (The text segment begins here)

0x1000-0000 = 268435456 = 2^28 = 1/16  = 1 inch     (The $gp data segment begins here)

0x10010000 = 268500992 = 2^28 + 2^16  = 1/16 + 1/65536 = (1 + 1/65536) inch (non-$gp data seg)

0x8000-0000 = 2147483648 = 2^31        = 1/2  = 8 inches  (Kernel text begins here)

0x9000-0000 = 2415919104 = 2^31 + 2^28 = 9 * 2^28 = 9/16  = 9 inches  (Kernel data begins here)

0xFFFF-0000 = 4294901760 = 2^32 - 2^16 = (15 + 65535/65536) inches  (Memory-mapped IO space)

0x1-0000-0000 = 4294967296 = 2^32      =  16 inches    This is the very top of address space

The code below verifies these assertions as documented in the comments with each line.
We use hyphens to break long hexadecimal strings into groups of 4 digits.
"""

# 2^22 = 0x0040-0000 and 1024 times that is 0x1-0000-0000, which is 2^32
print("{:x}     {:x}".format(pow(2, 22), 1024 * pow(2, 22)))

# 2^28 = 0x1000-0000 and 16 times that is 0x1-0000-0000, which is 2^32
print("{:x}     {:x}".format(pow(2,28), 16 * pow(2,28)))

# 2^28 + 2^16 = 0x10010000 and 65536 * 2^16 = 2^32
print("{:x}    {:x}".format(pow(2,28) + pow(2,16), 65536 * pow(2, 16)))

# 2^31 = 0x8000-0000
print("{:x}".format(pow(2,31)))

# 2^31 + 2^28 = 0x9000-0000
print("{:x}".format(pow(2,31) + pow(2,28)))

# 2^32 - 2^16 = 0xFFFF-0000
print("{:x}".format(pow(2, 32) - pow(2, 16)))
