import tkinter
from tkinter import ttk
import jtk
import logging

WIDTH = 10
class LabelRow(ttk.Frame):
    def __init__(self, parent, width=WIDTH):
        super().__init__(parent, padding="3 3 3 3")
        self.values = []
        for i in range(WIDTH):
            self.values.append(jtk.Label(self, width=1))
            self.values[i].set(WIDTH-1-i)
            self.values[i].grid(column=i, row=0, padx=1, pady=0)
        dummy = jtk.Label(self, width=WIDTH)
        dummy.grid(column=WIDTH, row=0, padx=20, pady=0)
class AndOrRow(ttk.Frame):
    def __init__(self, parent, count=WIDTH, readonly=False):
        super().__init__(parent, padding="3 3 3 3")
        
        self.values = []
        for i in range(count):
            self.values.append(jtk.BitField(self, self.master.doBad, readonly=readonly,
                                        postOK=self.postBitsOK))
            self.values[i].grid(column=i, row=0)#, padx=0, pady=0)
        self.total = jtk.RTVEntry(self, self.hexResultOK, self.master.doBad, width=WIDTH)
        self.total.unvalidatedSetStr('0')
        self.total.grid(column=WIDTH, row=0, padx=20, pady=0)
        if readonly:
            self.total.state(["readonly"])
            self.total.configure(validate='none')

    def postBitsOK(self):
        self.num = self.getBits()
        self.setHex(self.num)
        self.master.updateViews()

    def hexResultOK(self, string, delta):
        logging.debug("deciding on '{}'".format(string))
        try:
            result = int(string, 16) ## can we parse as hex integer??
        except:
            logging.debug("bad number format")
            return False, None
        else:  ## number format is ok; is value in range?
            logging.debug("got result {}".format(result))
            if 0 <= result < 1<<WIDTH:
                logging.debug("liking it")
                self.setBits(result)
                self.after(0, self.master.updateViews())
                return True, string
            else:
                logging.debug("not liking it")
                return False, None
        
    def getBits(self):  ## returns  integer
        val = 0
        for i in range(WIDTH):
            dig = self.values[i].get()
            val = 2*val + int(dig)
        return val

    def getHex(self):
        return int(self.total.get(), 16)

    def setBits(self, newValue):  ## newValue is integer
        assert 0 <= newValue < 1<< WIDTH
        logging.debug("setting to {}".format(newValue))
        for i in range(WIDTH):
            dig, newValue = newValue % 2, newValue // 2
            self.values[WIDTH-1-i].set(dig)

    def setHex(self, newValue):  ## newValue is integer
        assert 0 <= newValue < 1<< WIDTH
        logging.debug("setting to {}".format(newValue))
        self.total.unvalidatedSetStr(jtk.radStr(newValue, 16))

    def state(self, newState):
        for i in range(WIDTH):
            self.values[i].state(newState)
        self.total.state(newState)

class App(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)
        logging.basicConfig(level=logging.WARNING)
        self.lbls = LabelRow(self)
        self.lbls.pack(padx=0)
        self.aom1 = AndOrRow(self)
        self.aom2 = AndOrRow(self)
        self.aom3 = AndOrRow(self, readonly=True)
        self.aom1.pack(padx=10)
        self.aom2.pack(padx=10)
        ttk.Separator(self, orient="horizontal").pack(fill="x", padx=10)
        self.aom3.pack(padx=10)
        
        self.op = tkinter.StringVar()
        andButton = ttk.Radiobutton(self, text="and", variable=self.op,
                                    value='and', command=self.updateViews)
        orButton = ttk.Radiobutton(self, text="or", variable=self.op,
                                   value='or', command=self.updateViews)
        self.op.set('and')
        andButton.pack(side="left", padx=(100, 10))
        orButton.pack(side="right", padx=(10, 100))

    def updateViews(self):
        v1 = self.aom1.getHex()
        v2 = self.aom2.getHex()
        op = self.op.get()
        self.aom3.state(["!readonly"])
        if op == 'and':
            self.aom3.setBits(v1 & v2)
            self.aom3.setHex(v1 & v2)
        elif op == 'or':
            self.aom3.setBits(v1 | v2)
            self.aom3.setHex(v1 | v2)
        else:
            print("'{}' should be in ['and', 'or']".format(op))
        self.aom3.state(["readonly"])

    def doBad(self):
        self.bell()

if __name__ == "__main__":
    root = tkinter.Tk()
    root.title("Ands and Ors")
    a = App(root)
    a.pack()
    root.mainloop()
