import tkinter
from tkinter import ttk
import jtk
import model
import buttonpanel
import valuepanel
import logging

NUMROWS = 5
MAXMODULUS = 1<<32

class BaseChanger(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent, padding="10 10 10 10")
        self.myModel = model.Model(self, MAXMODULUS)
        
        baseValues = ['']
        for i in range(2,37): ## gives us 2-36
            baseValues.append(i)
        self.theVPs = []
        vp = valuepanel.TopValuePanel(self, baseValues, self.myModel)
        vp.pack(pady="5")
        self.theVPs.append(vp)
        for i in range(1, NUMROWS):
            vp = valuepanel.ValuePanel(self, i, baseValues, self.myModel)
            vp.pack(pady="5")
            self.theVPs.append(vp)

        self.buttonPanel = buttonpanel.ButtonPanel(self,self.myModel)
        
        self.buttonPanel.pack(pady="15")
        self.myModel.pack()
        ## default behavior on <Down> was messing up our behavior (incr/decr)
        self.bind_class("TCombobox", "<Down>", lambda e : None)

    def updateViews(self, skip):
        if skip is None:  ## for logging
            sk = "None"
        else:
            sk = skip.index
        if self.myModel.isGoodValue():
            value = self.myModel.getValue()
        else:
            value = None
        logging.info("updating to {} skipping {}".format(value, sk))
        for vp in self.theVPs:
            if vp != skip:
                vp.setNum(value)

    def updateDigits(self):
        logging.info("updating digits")
        for vp in self.theVPs:
            vp.setDigits()

    def doBad(self):
        if self.buttonPanel.bellButton.isSelected():
            self.bell()

    def someSigned(self):
        for vp in self.theVPs:
            if vp.signButton.instate(["selected"]):
                return True
        return False
                
if __name__ == "__main__":            
    root = tkinter.Tk()
    root.title("Base Changer")
    root.geometry('-20+40') ## 300 X 280 pixels; 20 from right side 40 down
    root.columnconfigure(0, weight=1)
    root.rowconfigure(0, weight=1)
    logging.basicConfig(level=logging.ERROR)
    logging.info("MAXMODULUS is {}".format(MAXMODULUS))

    bc = BaseChanger(root)
    bc.pack()

    root.mainloop()
