__author__ = 'jimmy'
import tkinter
from tkinter import ttk
import logging
import jtk

class Blinker:

    counting = True

    def __init__(self, blinkingThing, blinkRate=250):
        self.blinkingThing = blinkingThing
        if isinstance(blinkingThing, (list, tuple)):
            self.parent = blinkingThing[0].master
            ## assert that blinkingThing[0] and blinkingThing[1] have the correct types
        else:
            self.parent = blinkingThing.master
        self.blinkRate = blinkRate
        self.showing = True
        self.blinkCount = None

    # blinkingThing can be a Label or a pair: (a Canvas, id of a text object in the canvas)
    def blink(self, count = None, next=None):
        assert count is None or count > 0, "Blink count must be positive or None"
        self.blinkCount = count
        self.value = self.get()
        self.action = self.parent.after(self.blinkRate, self.keepBlinking, next)

    def keepBlinking(self: object, next):
        if self.blinkCount is None or self.blinkCount > 0:               ## keep blinking
            self.showing = not self.showing                              ## switch state
            if self.showing:
                self.set(self.value)
                if self.blinkCount is not None and Blinker.counting:
                    self.blinkCount -= 1                ## decrement when showing is ON so it stays on when we stop
            else:
                self.set("   ")
            self.action = self.parent.after(self.blinkRate, self.keepBlinking, next)        ## schedule another cycle
        else:
            if next is not None:
                next()

    def stop(self):
        if hasattr(self, "action"):
            self.parent.after_cancel(self.action)
            self.set(self.value)
            self.showing = True
            self.blinkCount = None

    def get(self):
        if isinstance(self.blinkingThing, (list, tuple)):                        ## our blinker is text in a Canvas
            return self.blinkingThing[0].itemcget(self.blinkingThing[1], "text")
        else:                                                                    ## our blinker is a Widget
            return self.blinkingThing.get()

    def set(self, newValue):
        if isinstance(self.blinkingThing, (list, tuple)):                        ## our blinker is text in a Canvas
            self.blinkingThing[0].itemconfigure(self.blinkingThing[1], text=newValue)
        else:                                                                    ## our blinker is a Widget
            self.blinkingThing.set(newValue)





if __name__ == "__main__":


    class BlinkerApp(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            self.win = win = tkinter.Canvas(self, width=300, height=100)
            win.pack()

            t1 = win.create_text(75, 50, text="abc")
            self.bct1 = Blinker((win, t1), 500)
            self.bct1.blink(5)

            t2 = win.create_text(125, 50, text="xyz")
            self.bct2 = Blinker((win, t2), 400)
            self.bct2.blink()

            self.l = jtk.Label(self)
            self.l.set(14)
            self.l.pack()
            self.bl = Blinker(self.l, 300)
            self.bl.blink()

            b1 = ttk.Button(self, text="stop", command=self.stopit)
            b1.pack()

            b2 = ttk.Button(self, text="pause", command=self.pause)
            b2.pack()

        def stopit(self):
            self.bct1.stop()
            self.bct2.stop()
            self.bl.stop()

        def pause(self):
            Blinker.counting = not Blinker.counting

    ## end of class BlinkerApp

    def main() -> object:
        root = tkinter.Tk()
        root.title("Model Register File")
        ra = BlinkerApp(root)
        ra.pack()
        root.mainloop()

    main()

