import tkinter
from tkinter import ttk
import logging
import jtk
import random
from crawler import Crawler
from registers import RegisterFile
from blinker import Blinker

"""
Model the register file depicted in chapter 4 of Patterson and Hennessy.
There are 8 registers, we use Comboboxes to pick register numbers, and
we use a Combobox to pick the value for the data bus. Use crawlers and
blinkers to highlight what is going on with registers in the cpu.
"""



if __name__ == "__main__":
    TopLeft_X = 400
    TopLeft_Y = 100
    DIM = 40
    FAT = 2*DIM        ## height and width of the crawling feature and width of fat lines
    SKINNY = DIM


    class RegisterApp(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            self.win = win = tkinter.Canvas(self, width=1000, height=700)
            win.pack()
            
            self.r = r = RegisterFile(win, TopLeft_X, TopLeft_Y, True, self.resetw)       ## create RegisterFile objack
            cbvalues = r.getComboValues()                         ## list of labels for the Comboboxes
            
            
    ## read data 1
            outputObject = jtk.Label(win)
            line = self.win.create_line(r.getRightX(), r.getRightY(0), r.getRightX()+300, r.getRightY(0),
                                             width=FAT, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.rd1crl = Crawler(self.win, line, outputObject)
            r.setRD1(self.rd1crl)

    ## read data 2            
            outputObject = jtk.Label(win)
            line = self.win.create_line(r.getRightX(), r.getRightY(1), r.getRightX()+300, r.getRightY(1),
                                             width=FAT, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.rd2crl = Crawler(self.win, line, outputObject)
            r.setRD2(self.rd2crl)

    ## read reg 1
            self.cmb1 = ttk.Combobox(self, values=cbvalues, width=3,
                                        state='readonly', takefocus=False)
            self.cmb1.bind("<<ComboboxSelected>>", self.crawl1)   ## crawl when Combobox changes
            self.cmb1.set(random.randint(0, len(cbvalues) - 1))
            line = self.win.create_line(r.getLeftX()-300, r.getLeftY(0), r.getLeftX(), r.getLeftY(0),
                                             width=SKINNY, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.rr1crl = Crawler(self.win, line, self.cmb1)
            r.setRR1(self.rr1crl)
            r.doRD1()                               ## make it start out showing the right value
            
    ## read reg 2
            self.cmb2 = ttk.Combobox(self, values=cbvalues, width=3,
                                        state='readonly', takefocus=False)
            self.cmb2.bind("<<ComboboxSelected>>", self.crawl2)
            self.cmb2.set(random.randint(0, len(cbvalues) - 1))
            line = self.win.create_line(r.getLeftX()-300, r.getLeftY(1), r.getLeftX(), r.getLeftY(1),
                                             width=SKINNY, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.rr2crl = Crawler(self.win, line, self.cmb2)
            r.setRR2(self.rr2crl)
            r.doRD2()                               ## make it start out showing the right value

    ## write reg           
            self.cmbwr = ttk.Combobox(self, values=cbvalues, width=3,
                                        state='readonly', takefocus=False)
            self.cmbwr.bind("<<ComboboxSelected>>", self.crawlwr)
            self.cmbwr.set(random.randint(0, len(cbvalues) - 1))
            line = self.win.create_line(r.getLeftX()-300, r.getLeftY(2), r.getLeftX(), r.getLeftY(2),
                                             width=SKINNY, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.wrcrl = Crawler(self.win, line, self.cmbwr)
            r.setWR(self.wrcrl)

    ## write data
            self.cmbwd = ttk.Combobox(self, values=list(range(25)), width=3,
                                        state='readonly', takefocus=False)
            self.cmbwd.bind("<<ComboboxSelected>>", self.crawlwd)
            self.cmbwd.set(random.randint(0, len(cbvalues) - 1))
            line = self.win.create_line(r.getLeftX()-300, r.getLeftY(3), r.getLeftX(), r.getLeftY(3),
                                             width=FAT, joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.wdcrl = Crawler(self.win, line, self.cmbwd)
            r.setWD(self.wdcrl)

            self.b = ttk.Button(self, text="pause", command=self.doButton)
            self.b.pack()
            Crawler.going = True

        def doButton(self, event=None):
            Crawler.going = not Crawler.going
            Blinker.counting = not Blinker.counting
            if Crawler.going:
                self.b["text"] = "pause"
            else:
                self.b["text"] = "continue"

        def crawl1(self, event=None):
            self.rr1crl.start(self.blink11)

        def blink11(self):
            thereg = int(self.cmb1.get())
            self.r.regs[thereg].bl.blink(4, self.blink12)

        def blink12(self):
            thereg = int(self.cmb1.get())
            self.r.regs[thereg].bd.blink(4, self.out1)

        def out1(self):
            self.r.doRD1()
            self.rd1crl.start(self.reset1)

        def reset1(self):
            self.rr1crl.reset()
            self.rd1crl.reset()

        def crawl2(self, event=None):
            self.rr2crl.start(self.blink21)

        def blink21(self):
            thereg = int(self.cmb2.get())
            self.r.regs[thereg].bl.blink(4, self.blink22)

        def blink22(self):
            thereg = int(self.cmb2.get())
            self.r.regs[thereg].bd.blink(4, self.out2)

        def out2(self):
            self.r.doRD2()
            self.rd2crl.start(self.reset2)

        def reset2(self):
            self.rr2crl.reset()
            self.rd2crl.reset()           

        def crawlwr(self, event=None):
            self.wrcrl.start(self.blinkwr1)

        def blinkwr1(self):
            thereg = int(self.cmbwr.get())
            self.r.regs[thereg].bl.blink()

        def blinkwr2(self):
            thereg = int(self.cmbwr.get())
            self.r.regs[thereg].bd.blink(None)

        def crawlwd(self, event=None):
            self.wdcrl.start()

        def resetw(self):
            thereg = int(self.cmbwr.get())
            self.r.regs[thereg].bl.stop()
            self.r.regs[thereg].bd.blink(4)
            self.wrcrl.reset()
            self.wdcrl.reset()

## end of class RegisterApp



    def main():
        root = tkinter.Tk()
        root.title("Model Register File")
        ra = RegisterApp(root)
        ra.pack()
        root.mainloop()
        
    main()


