import tkinter
from tkinter import ttk
from tkinter import font             # This solves a problem with Python 3.2
import logging
import string
#import sys
         
DISTINCT_COLORS = ['yellow', 'salmon', 'palegreen', 'hotpink',
          'lemonchiffon', 'darkorange', 'SeaGreen2', 'plum1',
          'gold', 'orange', 'chartreuse', 'orchid1',
          'khaki', 'tomato', 'springgreen', 'RosyBrown1']
	
## brief
##def baseN(num,b,numerals="0123456789abcdefghijklmnopqrstuvwxyz"):
##    return ((num == 0) and numerals[0]) or (baseN(num // b, b, numerals).lstrip(numerals[0]) + numerals[num % b]
"""Inverse function to int(str,radix) and long(str,radix)."""
def radStr(number, radix, width=1, digits=string.digits + string.ascii_uppercase):
   digitCount = len(digits)
   assert 2 <= radix <= digitCount
   result = []
   # convert some globals to locals for speed
   _divmod, _digits = divmod, digits
   append = result.append
   # check for a negative number
   if number < 0:
       number = -number
       isNegative = True
   else:
       isNegative = False
   # append digits in reverse order
   while number:
      (number, remainder) = _divmod(number, radix)
      append(_digits[remainder])
   # append leading 0s
   leading0s = digits[0] * (width-len(result)) # note: "0" * -i == ''
   append(leading0s)
   # append sign if necessary and reverse
   if isNegative:
      append('-')
   result.reverse()
   return ''.join(result)

def hexStr(number, width=1):        ## returns string representing a number in hexadecimal
   return '0x' + radStr(number, 16, width)

def dividedHexStr(num):             ## e.g.   0xffff-ffe0
    str = hexStr(num, 8)
    return str[:6]+'-'+str[6:]


"""
Here's a ttk.Checkbutton that's easier to use.
"""
class Checkbutton(ttk.Checkbutton):
    def __init__(self, parent, initiallySelected, **kwargs):
        ttk.Checkbutton.__init__(self, parent, **kwargs)
        self.set(initiallySelected)

    def isSelected(self):
        return self.instate(["selected"])

    def set(self, willBeSelected):
        if willBeSelected:
            self.state(["selected", "!alternate"])
        else:
            self.state(["!selected", "!alternate"])
        callback = self['command']
        logging.debug("checkbutton callback is: {}".format(callback))

    def get(self):
        return self.isSelected()

"""
Here's a ttk.Label that knows get and set.
"""
class Label(ttk.Label):
    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        self.value = self["text"]
        self.visible = True

    def get(self):
        return self.value

    def set(self, newValue):
        self.value = newValue
        self.setText()

    def setText(self):
        if self.visible:
            self["text"] = self.value
        else:
            num = self["width"]
            try: num = int(num)
            except ValueError as ex: num = 2
            self["text"] = " " * num

    def setVisible(self, willBeVisible):
        self.visible = willBeVisible
        self.setText()

"""
Here's a ttk.Entry that knows get and set.
"""

class Entry(ttk.Entry):
    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)

    def set(self, newValue):              ## newValue is string
        self.delete(0, "end")
        self.insert(0, newValue)


"""
Draw a line on the canvas and add a text component to the line at one end
or at the center. The endpoints of the line are (x1, y1) and (x2, y2). The
program will make sure that x1 <= x2, which simplifies the programming.
(Whichever way the line slopes, (x1, y1) will be at its left end.) The
given anchor (one of "neswc") will determine where the text component is
placed. If the anchor is "c", the text component will be in the center of the
line; otherwise it will be at one end. In the other four cases the line
will touch the component at the side of the component specified by the anchor.

There are two factory methods that use two points or a point and a length to
determine the endpoints of the line.

If the component is None, the line will be drawn as a plain line. In this case,
if the line is constructed from point-length, the anchor will determine which
direction the line takes from the parameter point. Think of the other end as
attaching to the end of an imaginary text component at the side of the component
given by the anchor. If the anchor is one of "ns", the line will be vertical;
if it is in "ew", the line will be horizontal.
E.g. ptLen(canvas, pt, length, width, "e", None, irrelevant)
will draw a horizontal line westward from pt so the free end will attach to the
east side of the imaginary text component.

The text components that TextLine is intended to work with are jtk.RTVEntry,
jtk.Label, and ttk.Combobox. The jtk.RTVEntry will do all its usual validation
stuff in this situation provided that it is given an appropriate resultOK
function. A jtk.Label will implement setVisible(willBeVisible). ttk.Button
also works just fine. (memory.py makes a good example)

In TextLine.__init__() the parameters are as follows: canvas is the tkinter.Canvas
that will do the drawing; x1, y1 x2, y2 are the coordinates of the two ends
of the line; width is the line's thickness in pixels; anch is one of "neswc"
and determines which side of the text component is touched by the line; item
is the text component; and itemPixels is the size in pixels of the window in the
canvas that the item will be placed into. Here are some examples of typical
items that can be used with TextLine along with comments documenting some of
their methods that can be used to set and retrieve their contents:


        self.val = jtk.RTVEntry(self, ok, bad, width=25, justify="center")
##        inherits get() from ttk.RTVEntry and set(newString) and
##        unvalidatedSetStr(newString) from jtk.RTVEntry

        self.val = jtk.Label(self, text=initial)
##        self.val.get()             ## getter function
##        self.val.set(newText)      ## setter function
##        self.val.setVisible(False)

        self.op = ttk.Combobox(self, values=['0', '1', '2', '6', '7'], width=3,
                                    state='readonly', takefocus=False)
        self.val.bind("<<ComboboxSelected>>", self.doOp)
##        self.val.get()
##        self.val.set(newString)

        self.tex = canvas.create_text(x, y, text="some text", anchor='e')
##        can be set using delete and insert, but apparently can't be read.

"""


class TextLine:
    def __init__(self, canvas, x1, y1, x2, y2, width, anch, textItem, itemPixels):
        assert anch in "neswc", "'{}' must be in 'neswc'".format(anch)
        if x1 > x2:                     ## make sure that x1 <= x2
            x1, x2 = x2, x1
            y1, y2 = y2, y1             ## be sure to swap BOTH coordinates
    ## wx, wy will be the coordinates of the item's window that connects to the line
        if anch == "n":             ## connect item to bottom end of line
            if y1 >= y2:
                wx, wy = x1, y1
            else:
                wx, wy = x2, y2
        elif anch == "e":           ## connect item to left end of line
            wx, wy = x1, y1
        elif anch == "s":           ## connect item to top end of line
            if y1 <= y2:
                wx, wy = x1, y1
            else:
                wx, wy = x2, y2
        elif anch == "w":           ## connect item to right end of line
            wx, wy = x2, y2
        elif anch == "c":           ## place item over center of line
            wx, wy = (x1+x2)/2, (y1+y2)/2
        else:                       ## the assert won't let us get here
            pass
        
        canvas.create_line(x1, y1, x2, y2, width=width)
        if textItem is not None:
            canvas.create_window(wx, wy, window=textItem, anchor=anch, width=itemPixels)
            self.textItem = textItem
            out = ""
        else:
            out = "out"
        logging.debug("with{} text item {} {} {} {}".format(out, x1, y1, x2, y2))

    def get(self):
        assert self.textItem is not None
        return self.textItem.get()

    def set(self, text):
        assert self.textItem is not None
        self.textItem.set(text)

    def setVisible(self, willBeVisible):
        assert isinstance(self.textItem, Label)
        self.textItem.setVisible(willBeVisible)  ## throws exception except for Label
        
## factory methods called as TextLine.endPoints, TextLine.ptLen.
    @classmethod
    def endPoints(cls, canvas, p1, p2, width, anch, item, itemPixels):
        x1, y1 = p1[0], p1[1]
        x2, y2 = p2[0], p2[1]
        return cls(canvas, x1, y1, x2, y2, width, anch, item, itemPixels)

    @classmethod
    def ptLen(cls, canvas, point, length, width, anch, item, itemPixels):
        assert anch in "nesw"               ## "c" wouldn't give direction
        x1, y1 = point[0], point[1]
        if anch in 'es':
            length = - length               ## move left if anchor on left end, up if on south end
        if anch in 'ew':
            x2, y2 = x1+length, y1
        elif anch in 'ns':
            x2, y2 = x1, y1+length
        else:
            pass                           ## can't get here
        return cls(canvas, x1, y1, x2, y2, width, anch, item, itemPixels)

"""
RTVEntry is a RealTimeValidatedEntry; it follows the typing in real time and
disallows changes that make the displayed text unsuitable. This looks
better than allowing the change to take place and then backing it out.

The resultOK function should be tailored to the particular application.
It's signature is resultOK(self, string) -> (Boolean, string). The Boolean
says whether the parameter string passed the test. The returned string is
what will be stuffed into the RTVEntry, which allows us to change what is
displayed. This is useful, for instance, if we want to display upper case
even if lower case was typed.

The onBad parameter may be used to specify a custom method to be called after
each incorrect keystroke. If this parameter is omitted, no action will occur
beyond leaving the contents of the RTVEntry unchanged. If this parameter specifies
None, each incorrect keystroke will call the built-in bell() method in
addition to ignoring the keystroke.

Doing what we want is made more difficult by the way tkinter validation works.
Here's the problem: if there is a selection in the RTVEntry, and the user
types a character or pastes in text, the validate function gets called twice,
a delete (for the selection) followed by an insert (for the typed or pasted
character(s)). What's wrong with that is that the change can be rejected only
by the insert validator, and by the time it gets control, the delete has already
taken place, so the original string (from before the delete) will have to be
restored. This will be visible to the user and is ugly.
 
Here's how we get around that. In the first place, we never let the ordinary
course of events change the text in the widget. We force in the correct value
every time. This requires this validator to return False every time.

WARNING!!!!! NEVER let a validate function return None. That will turn validating
OFF!!!. This will produce mysterious symptoms and cause you no end of trouble.

There are three possible actions the user can take that require validation:
A. a delete action: typing a delete key or doing a Cut
B. an insert action: typing a character key or doing a Paste
C. a replace action: performing an insert action when there is a selection. This
   results in two calls to the validator function: a delete of the selection
   followed by an insert at the cursor point left by the delete. (A delete action
   when there is a selection falls into case A since there is nothing to insert.
   The only difference between this and ordinary case A is that several characters
   will be deleted if the selection has length > 1).

It is the two calls to the validator in case C that cause the problem mentioned above.

MOUNT SOAPBOX
The RIGHT way for the widget to handle validation is to make ONE call to the validate
method for each user action. A replace method will handle deletes and insertions as
subcases: delete is when the replacing text is empty, insert when the replaced text
is empty. The SWT does this. Java Swing, does not. tkinter does not, but it does
provide the new strings to the validation method, which facilitates the accept/reject
decision. I wish somebody would wrap the SWT for Python. I think that fixing tkinter
would require fixing Tk, which is out of the control of the Python community.
DISMOUNT SOAPBOX

There are four cases for the validator to distinguish:
1. delete caused by case A above
2. delete caused by case C above
3. insert caused by case C above
4. insert caused by case B above

In cases 1 and 4 we just force in the validator's result string. The only difference
between them is that in case 4 we move the cursor past the new text.
In case 2 we store the result string and insertion point but don't delete anything
In case 3 we retrieve the stored string and insertion point and use them along
with the inserted string to construct the result of the two actions combined.
If the user types a delete key when there is a selection, that fits cases 1 and 2.
In this case there is no following insert, so this must be taken as case 1. Otherwise,
the delete results will simply be placed in temporary storage and not in the widget.

The cases are distinguished as follows:
All delete keys and Cut actions are bound to handler methods that set a flag BEFORE
the validate for the delete happens. The validate for the delete looks at this flag
to distinguish case 1 from case 2. Thus in the ambiguous case metioned above the
validating code for deletes will find the flag set and properly handle it as case 1.
In case 2 the code stores the string and the cursor position that would result from
the delete but does not change the widget. The validation code for inserts looks at
these variables to distinguish cases 3 and 4. All these variables are reset at the
end of each operation when the new string is forced into the widget.

There is one other tricky possibility: if the user takes a delete action with an empty
string in the widget, the flag will be set, but the validator will not be called, so the
flag is not reset. This would be a problem if the next user action caused a selection
to be replaced (case 2/3); the results of the accompanying delete would not be stored,
and the insert would begin with an incorrect string to insert into. However, no
selection is possible in an empty widget; as long as the user keeps taking delete
actions, nothing else happens. The only action possible at this point that will call
the validator is a plain insert (case 4), which will reset the flag so no harm is done.

If I had known what I now know about tkinter, this would have been a piece of cake.
As it was, it was a pain in the ass, but I learned a lot, some of it TMI.
"""

## The onBad parameter specifies how the RTVEntry reacts to incorrect typing. There are three possibilities:
##    the programmer specifies a method; it will be called when the user types a character that fails the test
##    the programmer specifies, "None"; any failing characters are simply ignored
##    the programmer omits this parameter; any failing character will ring the bell
## The default parameter "default" is there because default None can't be distinguished from programmer-specified None
class RTVEntry(ttk.Entry):
    def __init__(self, parent, resultOK, onBad="default", **kw):
        super().__init__(parent, validate="key", **kw)
        self.resultOK = resultOK
        if onBad is "default":
           self.onBad = self.bell
        else:
           self.onBad = onBad
#        print("self.onBad is: ", self.onBad)
        vcmd = (self.register(self.OnValidate), 
                '%d', '%i', '%P', '%S')
        self.configure(validatecommand=vcmd)
        self.justDelete = False  ## True if a delete operation was typed.
        self.doneDelete = None   ## String the delete would have left.
        self.deleteIndex = None  ## Insertion point the delete would have left.
        self.bind('<BackSpace>', self.onDelete)  ## catch
        self.bind('<Delete>', self.onDelete)     ## delete
        self.bind('<<Cut>>', self.onDelete)      ## operations

    def onDelete(self, event):
        logging.debug("onDelete found justDelete = {}".format(self.justDelete))
        self.justDelete = True
        
    def OnValidate(self, action, index, new, delta):
        self.displayValidate(action, index, new, delta)
        action = int(action)
        index = int(index)
        val = None
        if action == 0:     ### this is a delete
            if self.justDelete:  ## case 1: plain delete
                val = self.testAndSet(new, delta, index)
                logging.debug("plain delete sez val is '{}'".format(val))
            else:                ## case 2: replace; so delay the delete
                self.doneDelete = new     ## remember result of delete
                self.deleteIndex = index  ## and where following insert will go
                logging.debug("delayed delete storing '{}' and {}".format(new, index))
            
        elif action == 1:        ### this is an insert
            if self.deleteIndex is None:      ## case 4: plain insert
                val = self.testAndSet(new, delta, index + len(delta))
                logging.debug("plain insert sez val is '{}'".format(val))
            else:                             ## case 3:there is a pending delete
                resultString = (self.doneDelete[:self.deleteIndex] + delta +
                                    self.doneDelete[self.deleteIndex:])
                val = self.testAndSet(resultString, delta, self.deleteIndex + len(delta))
                logging.debug("replace sez val is '{}'".format(val))
        return False
    """
    We always force the correct value back into the RTVEntry, so we don't want the
    usual process to continue. Returning False here (the line above) will keep
    that from happening. This has another consequences: we can't set the
    invalidcommand option for our target because it will be used every time
    since we always return False.Therefore, we have to call onBad directly
    in testAndSet().
    """


    """
    Try to set value to string and cursor to index. Whether this actually happens
    depends on the values returned by resultOK(string). Generally, resultOK()
    returns (Boolean, string). If the Boolean is False, we make no change; if it is
    True, we change our contents to the string. Notice that the string that comes
    back from resultOK() need not be the same as the one we sent to it. This allows
    us, for instance, always to display upper case even if the user types lower case.
    If the tuple returned by resultOK() has a third element, it will be used as the
    new position of the cursor instead of our parameter "index". This was added to the
    code to allow us to leave the cursor in the correct position in a case where deleting
    the last character in a digit string returned "0" instead of the empty string.
    """
    def testAndSet(self, new, delta, index):
        logging.debug("forcing in '{}' leaving cursor at {}".format(new, index))
        retval = self.resultOK(new, delta)         ## will be a tuple
        ok, result = retval[0], retval[1]   ## these two are required
        if len(retval) > 2:                 ## if there's a third element in the tuple
            index = retval[2]               ## use it to override our parameter "index"
            logging.debug("changing cursor to {}".format(index))
        if ok:
            ## We gung stuff result into the RTVEntry, so it must be a string.
            ## We could just stuff in our parameter string, but this way allows
            ## resultOK() to change what gets stuffed; sometimes that's useful
            ## E.g. maybe we display uppercase even it lowercase is typed.
            assert isinstance(result, str), "testAndSet needed string, got {}".format(type(result))
            self.delete(0, "end")
            self.insert(0, result)
            self.icursor(index)
        else:
            if self.onBad is not None:
                self.onBad()
                result = None
        self.resetFlags()
        return result  ## used for logging in OnValidate()
        
    def resetFlags(self):
        self.justDelete = False
        self.deleteIndex = None
        self.doneDelete = None
        
    def displayValidate(self, action, index, new, delta):
        if action == '0':
            fmtStr = "would delete '{}' at {} yielding '{}'----------"
        elif action == '1':
            fmtStr = "would insert '{}' at {} yielding '{}'----------"
        else:
            fmtStr = "otheraction of '{}' at {} yielding '{}'----------"
        logging.debug(fmtStr.format(delta, index, new))

    """
    All right now listen up. Any change of the text in an RTVEntry will
    cause validation to be called first. This includes directly stuffing
    a new string in without any typing having been done. This feature is
    turned off during validation to forestall recursive calls to the
    validator, which would cause infinite recursion. This next function
    can be called from outside the validator, and it will cause validation
    to take place. The one after it, unvalidatedSetStr(), can be called from
    outside validation in cases where we know validation is unnecessary. It
    turns validation off at the beginning and turns it back on as it exits.
    The unvalidated version is used several times in Valuepanel and once in
    BaseChanger.
    """
    def setStr(self, newValue):              ## newValue is string
        self.delete(0, "end")
        self.insert(0, newValue)

    def unvalidatedSetStr(self, newValue):   ## newValue is string
        self.config(validate="none")
        self.delete(0, "end")
        self.insert(0, newValue)
        self.config(validate="key")

    def set(self, newValue):
        self.setStr(newValue)
"""
BitField is an RTVEntry that allows only 0 or 1 to be typed in it.
In this case we have a special resultOK method that allows
overtyping of the current entry. This requires the caller
to supply the character typed as well as the complete new string.
"""
class BitField(RTVEntry):
    def __init__(self, parent, doBad, initial='0', postOK=None, readonly=False):
        super().__init__(parent, self.resultOK, doBad, width=1)
        self.unvalidatedSetStr(initial)
        self.postOK = postOK
        if readonly:
            self.configure(validate='none')
            self.state(["readonly"])

## We want to enable overwriting in bit fields, so we look at the
## character typed (delta), not the complete new string (string).
## If we want to initiate some action after the edit
## takes place, we must do it in postOK, not here, since
## the edit won't be complete until after we return.
    def resultOK(self, string, delta):
        if delta == '0' or delta == '1':
            if self.postOK is not None:
                self.after(0, self.postOK)
            return True, delta        
        else:
            return False, None

        
"""use a ttk.TreeView as a multicolumn ListBox with an optional vertical scrollbar"""
class McListBox(ttk.Frame):
    def __init__(self, parent, headers, **kwargs):
        super().__init__(parent)
        self.headers = headers        
        # create a treeview
        self.tree = ttk.Treeview(self, columns=headers, show="headings", **kwargs)
        self.tree.grid(column=0, row=0, sticky='nsew', in_=self)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        for col in headers:
            self.tree.heading(col, text=col.title())
            # adjust the column's width to the header string
            self.tree.column(col, width=tkinter.font.Font().measure(col.title()))

    def verticalScroll(self):   ## add the scrollbar
        vsb = ttk.Scrollbar(self, orient="vertical", command=self.tree.yview)
        self.tree.configure(yscrollcommand=vsb.set)
        vsb.grid(column=1, row=0, sticky='ns', in_=self)
   
    def getSelectionId(self):
        id = self.tree.selection()[0]
##        print("got selection ID: ", id)
        return id

    def getSelectionValues(self):
        id = self.getSelectionId()
        return self.getById(id)

    def getById(self, id):
        retval = []
        for col in self.headers:
            retval.append(self.tree.set(id, col))
##        print("selection values: ", retval)
        return retval

    def setSelectionValue(self, col, val):
        sid = self.getSelectionId()
        self.tree.set(sid, col, val)

    def idFromIndex(self, index):
        id = "I" + radStr(index, 10, 3)
##        print("ID {} from index {}".format(id, index))
        assert self.indexFromId(id) == index
        return id

    def indexFromId(self, id):
        idx = id[1:]
        index = int(idx)
##        print("index {} ({}) from ID {}".format(index, self.tree.index(id), id))
        return index
    
    def getByIndex(self, index):        
        id = self.idFromIndex(index)
        return self.getById(id)

    def setByIndex(self, index, col, val):
        id = self.idFromIndex(index)
        self.tree.set(id, col, val)

    def selectByIndex(self, index):
        id = self.idFromIndex(index)
        self.tree.selection_set(id)


class ResizingCanvas(tkinter.Canvas):
    def __init__(self,parent,**kwargs):
        super().__init__(parent,**kwargs)
        self.bind("<Configure>", self.on_resize)
        self.height = self.winfo_reqheight()
        self.width = self.winfo_reqwidth()

    def on_resize(self,event):
        # determine the ratio of old width/height to new width/height
        wscale = float(event.width) / self.width
        hscale = float(event.height) / self.height
        self.width = event.width
        self.height = event.height
        # resize the canvas 
        self.config(width=self.width, height=self.height)
        # rescale all the objects tagged with the "all" tag
        self.scale("all",0,0,wscale,hscale)


if __name__ == "__main__":

    print("radStr(123, 10): ", radStr(123, 10))
    print("radStr(123, 10, 10): ", radStr(123, 10, 10))
    print("radStr(12, 2): ", radStr(12, 2))
    print("radStr(12, 2, 10): ", radStr(12, 2, 10))
    print("radStr(-13, 2): ", radStr(-13, 2))
    print("radStr(13, 2, 10): ", radStr(13, 2, 10))
    print("radStr(-13, 2,, 10): ", radStr(-13, 2, 10))
    print("radStr(37, 36, 10): ", radStr(37, 36, 10))
    print("radStr(35, 36, 10): ", radStr(35, 36, 10))

    ## This makes NO sense except as illustration of capabilities, which is
    ## what it's here for. We use "justify" when we create the jtk.RTVEntry
    ## just to show that options for ttk.RTVEntry still work in this subclass.
    ## Then it gets really stupid: our validating function, resultOK
    ## accepts strings that are no longer than 4 characters and changes
    ## them to uppercase, just to demonstrate that we can do that. Now we
    ## do a couple of setStr's to show that they trigger validation.
    ## We call setStr("good") to demonstrate that it accepts and does the
    ## case change. Then we call setStr("better") to demonstrate that it
    ## rejects and calls doBad. Finally doBad calls setStr("worst") to
    ## show that validation is not called in this case. This is because doBad was
    ## called by resultOK1, and resultOK1 was called by the validator, so doBad is
    ## part of the validation of "better". Since validation is disabled when
    ## validation is already in progress, "worst" is not validated. We can see
    ## this because it is accepted even though it is too long, it stays in lower
    ## case, and there are no logging printouts for it as there are for the others.
    
    class App(ttk.Frame):
        def __init__(self,parent):
            super().__init__(parent)
            self.parent = parent
            logging.basicConfig(level=logging.INFO)
            self.e1 = e1 = RTVEntry(self, self.resultOK1, self.doBad1,        ## supply an explicit onBad method
                        width=25, justify="right")                          ## RTVEntry options still work
            e1.pack(padx=35, pady=5)

            self.canvas = canvas = ResizingCanvas(self, width=700, height=400)
      # these are here to prevent a long sequence of resizings that happen for reasons that I don't understand
      # I made this a ResizingCanvas just to show that it wouldn't bomb
            parent.minsize(700, 400)
            parent.maxsize(700, 400)

            e2 = RTVEntry(self, self.resultOK2, width=25, justify="center")  ## use default bell() for onBad method
            e2.set("12")
            ##Here x1=650 is greater than x2=500; TextLine.__init__ still does the right thing.
            self.we2 = we2 = TextLine(canvas, 650, 235, 500, 235, 15,"w", e2, 35)
            
            e3 = RTVEntry(self, self.resultOK1, None, width=50, justify="right")  ## explicitly say no onBad method
            e3.set("abc")
            self.we3 = we3 = TextLine(canvas, 100, 300, 300, 300, 20, "c", e3, 60)
            
            self.cbox = cbox = ttk.Combobox(self, values=['0', '1', '2', '6', '7'],
                                        width=3, state='readonly', takefocus=False)
            cbox.bind("<<ComboboxSelected>>", self.doOp)
            cbox.set("2")
            self.wc = wc = TextLine(canvas, 50, 100, 150, 200, 10, "n", cbox, 40)

            self.cb = Checkbutton(self, True, text="check", command=self.doCheck)
            TextLine(canvas, 400, 300, 450, 300, 25, "w", self.cb, 70)

            self.b = ttk.Button(self, text="Toggle Checkbutton", command=self.doButton)
            TextLine(canvas, 500, 350, 600, 350, 20, "e", self.b, 175)

            l = Label(self, text="points")
            self.wl = TextLine.endPoints(canvas, (50, 50), (200, 50), 30, 'c', l, 50)
            
            ## the next two begin at points with the same X-coordinate and go in different directions
            q = Label(self, text="horizE")            
            TextLine.ptLen(canvas, (400, 50), 100, 15, 'e', q, 50)

            r = Label(self, text="horizW")            
            TextLine.ptLen(canvas, (400, 80), 100, 15, 'w', r, 50)
            
            ## the next two begin at points with the same Y-coordinate and go in different directions
            e4 = Entry(self)
            e4.set("vertN")
            TextLine.ptLen(canvas, (600, 100), 40, 15, 'n', e4, 50)
            
            self.t = Label(self, text="vertS")            
            self.tt = TextLine.ptLen(canvas, (650, 100), 40, 15, 's', self.t, 50)

            TextLine.ptLen(canvas, (310, 200), 80, 50, 'w', None, None) ## short fat line with no text item

            self.next = ttk.Button(self, text="next", command=self.doStuff1)
            canvas.create_window(10, 360, anchor='w', window=self.next)

            canvas.pack()

        def doStuff1(self):
            self.next.config(command=self.doStuff2)
            self.e1.setStr("good")    ## validates insert
            print("inserting 'good'")

        def doStuff2(self):
            self.next.config(command=self.doStuff3)
            print("inserting 'better', which is too long; you should hear a beep from validator")
            print("validator inserts 'worst', which is also too long, but is not validated since the validator inserted it")
            self.e1.setStr("better")  ## validates delete/insert and fails (too long)

        def doStuff3(self):             ## exercise generalized get
            self.next.config(command=self.doStuff4)
            print("we2 has: ", self.we2.get())
            print("wl has: ", self.wl.get())
            print("wc has: ", self.wc.get())

        def doStuff4(self):             ## exercise generalized set using numbers
            self.next.config(command=self.doStuff5)
            print("setting we2, wl, and wc using integers: 25, 56, 7")
            self.we2.set(25)
            self.wl.set(56)
            self.wc.set(7)

        def doStuff5(self):             ## exercise generalized set using strings
            self.next.config(command=self.doStuff6)
            print("setting we2, wl, and wc using strings: 78, xyz, 0")
            self.we2.set('78')
            self.wl.set('xyz')
            self.wc.set('0')
            self.tt.setVisible(False)
            print("tt has: ", self.tt.get(), " but you can't see it")

        def doStuff6(self):
            self.next.config(command=self.parent.destroy)    # doesn't work; why?
            self.tt.setVisible(True)
            print("now you can see tt again")
            print("next line will throw an exception")
            self.we2.setVisible(False)          ## throws exception
            

        def resultOK1(self, string, delta):
            return len(string) < 5, string.upper()

            
        def resultOK2(self, string, delta):
            if string == '':
                retval = True, '0', 1           ## third value in tuple positions cursor       
            else:
                try:
                    result = int(string)
                except:
                    return False, None          ## not changing, so bail out now
                else:
                    retval = 0 <= result <= 1000, self.trim(string)  ## normal two tuple
            self.after(0, self.doOp)                                 ## 0 ms seems to work ok
            return retval

        def trim(self, string):  ## get rid of leading zeroes but not entire string
                i = 0
                for i in range(len(string) - 1):
                    if string[i] != '0': break
                    i += 1
                return string[i:]

        def doOp(self, evt=None):
            pass
            op = self.cbox.get()
            op = int(op)
            if op == 0:     ## bitwise and
                print("combobox sez 0(and)")
            elif op == 1:   ## bitwise or
                print("combobox sez 1(or)")
            elif op == 2:   ## add
                 print("combobox sez 2(add)")
            elif op == 6:   ## subtract
                print("combobox sez 6(sub)")
            elif op == 7:   ## slt (set on less than)
                print("combobox sez 7(slt)")
            
        def doBad1(self):
            self.bell()
            self.e1.setStr("worst")       ## doesn't validate since called by validator

        def doCheck(self):
            print("Checkbutton here with {}".format(self.cb.isSelected()))
            self.wl.setVisible(self.cb.isSelected())

        def doButton(self):
             self.cb.set(not self.cb.isSelected())
## Checkbutton command method not called automatically when button state changed by program
#             self.doCheck()

    def main():
        root = tkinter.Tk()
        root.title("Playing with TextLines")
        a = App(root)
        a.pack()
        root.mainloop()

    main()
